use bevy::{prelude::*, render::texture::TextureFormat};
use mandelbrot::Mandelbrot;
use std::time::Instant;

fn main() {
    // Mandelbrot::escape_time_image();
    App::build()
        .add_plugins(DefaultPlugins)
        .add_startup_system(startup.system())
        .add_resource(SpriteScale::default())
        .add_system(mandelbrot_render_system.system())
        .add_system(transform_mandelbrot_system.system())
        .run();
}

struct SpriteScale {
    zoom: f64,
    horiz: f64,
    verti: f64,
    step: f64,
    iterations: u32,
}

impl Default for SpriteScale {
    fn default() -> Self {
        Self {
            zoom: 1.0,
            horiz: 0.0,
            verti: 0.0,
            step: 0.01,
            iterations: 64,
        }
    }
}

fn transform_mandelbrot_system(
    input: Res<Input<KeyCode>>,
    mut scale: ResMut<SpriteScale>,
    // camera: Res<Camera>,
) {
    let zoom = input.pressed(KeyCode::Q) as i8 - input.pressed(KeyCode::E) as i8;
    let horizontal = input.pressed(KeyCode::D) as i8 - input.pressed(KeyCode::A) as i8;
    let vertical = input.pressed(KeyCode::W) as i8 - input.pressed(KeyCode::S) as i8;
    let step_mod = input.pressed(KeyCode::R) as i8 - input.pressed(KeyCode::F) as i8;
    let iter_mod = input.pressed(KeyCode::T) as i8 - input.pressed(KeyCode::G) as i8;

    scale.step += (scale.step / 10.0) * step_mod as f64;
    scale.verti += scale.step * vertical as f64;
    scale.horiz += scale.step * horizontal as f64;
    scale.zoom += scale.step * zoom as f64;

    if iter_mod == 1 {
        scale.iterations += (scale.step * 100.0) as u32;

        if scale.iterations > 512 {
            scale.iterations = 512;
        }
    } else if iter_mod == -1 {
        let step = (scale.step * 100.0) as u32;
        let diff = scale.iterations as i64 - step as i64;

        if diff < 32 {
            scale.iterations = 32;
        } else {
            scale.iterations -= step;
        }
    }

    if scale.zoom < 0.0 {
        // We can't go below 0
        scale.zoom -= scale.step * zoom as f64;

        // slow down our step function
        scale.step /= 10.0;
    }
}

fn mandelbrot_render_system(
    materials: Res<Assets<ColorMaterial>>,
    mut textures: ResMut<Assets<Texture>>,
    scale: Res<SpriteScale>,
    mut query: Query<(&mut Mandelbrot, &Handle<ColorMaterial>)>,
) {
    for (mut fractal, handle) in query.iter_mut() {
        if let Some(material) = materials.get(handle) {
            if let Some(texture_handle) = &material.texture {
                if let Some(texture) = textures.get_mut(texture_handle) {
                    let z = scale.zoom;
                    let h = scale.horiz;
                    let v = scale.verti;
                    let iters = scale.iterations;

                    let start = Instant::now();
                    texture.data.copy_from_slice(fractal.generate_scaled_image(
                        ((-2.5 * z) + h, (1.0 * z) + h),
                        ((-1.0 * z) - v, (1.0 * z) - v),
                        iters,
                    ));
                    let _diff = Instant::now() - start;
                }
            }
        }
    }
}

fn startup(
    mut commands: Commands,
    mut textures: ResMut<Assets<Texture>>,
    mut materials: ResMut<Assets<ColorMaterial>>,
) {
    let img_size = Vec2::new(Mandelbrot::width() as f32, Mandelbrot::height() as f32);
    let img_buf = Mandelbrot::new().generate_image().to_vec();

    let texture_handle = textures.add(Texture::new(
        img_size,
        img_buf,
        TextureFormat::Rgba8UnormSrgb,
    ));

    commands
        .spawn(Camera2dComponents::default())
        .spawn(SpriteComponents {
            material: materials.add(texture_handle.into()),
            ..Default::default()
        })
        .with(Mandelbrot::new());
}
